from dataclasses import dataclass
from typing import Dict, Optional, Union


@dataclass
class Point:
    x: float
    y: float
    z: Optional[float] = None


@dataclass
class IntPoint:
    x: int
    y: int
    z: Optional[int] = None


CenterPosition = Union[Point, Dict[str, float]]



import logging
from pathlib import Path
from typing import Optional, List, Union

import numpy as np
import pydicom
from pydicom import FileDataset

from .dicom_series import DicomSeries
from ..helpers.check_path_is_valid import check_path_is_valid_path
from ..helpers.pixel_data import get_pixel_array
from ..helpers.voxel_data import VoxelData

logger = logging.getLogger(__name__)


class DoseMatrix(DicomSeries):
    """A class to manage dose matrix series from Treatment planning systems.

    """
    def __init__(self, file: Path, dcm: Optional[FileDataset] = None):
        if dcm is None:
            dcm = pydicom.dcmread(fp=file, stop_before_pixels=True)

        if "SeriesInstanceUID" not in dcm:
            raise ValueError("The DICOM file does not contain a series instance UID")

        super().__init__(series_instance_uid=dcm.SeriesInstanceUID)
        self.DoseGridScaling: Optional[List[Optional[float]]] = []
        self.DoseSummationType: Optional[List[Optional[str]]] = []
        self.DoseType: Optional[List[Optional[str]]] = []
        self.DoseUnit: Optional[List[Optional[str]]] = []
        self.ImageVolume: Optional[List[np.ndarray]] = []
        self.Origin: Optional[List[List[float]]] = []

        self.add_file(file=file, dcm=dcm)

    def add_file(self, file: Union[Path, str], dcm: Optional[FileDataset] = None):
        """Add a file to the objects list of files

        First performs a check that the file path is valid and that is has the same Series Instance UID as the class
        object

        Args:
            file: Path to where the file to be added is stored on disc
            dcm: The DICOM-file imported to a FileDataset object

        Raises:
            TypeError: If the file is not a valid path
            ValueError: if SeriesInstanceUID of the file is not the same as the SeriesInstanceUid attribute
        """
        file = check_path_is_valid_path(file)

        if any([True if obj == file else False for obj in self.FilePaths]):
            # Return None since the file is already in the volume
            return

        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if dcm.SeriesInstanceUID != self.SeriesInstanceUid:
            msg = f"Wrong SeriesInstanceUID. Expected: {self.SeriesInstanceUid}; Input: {dcm.SeriesInstanceUID}"
            raise ValueError(msg)

        self.DoseGridScaling.append(float(dcm.DoseGridScaling) if "DoseGridScaling" in dcm else None)
        self.DoseSummationType.append(dcm.DoseSummationType if "DoseSummationType" in dcm else None)
        self.DoseType.append(dcm.DoseType if "DoseType" in dcm else None)
        self.DoseUnit.append(dcm.DoseUnit if "DoseUnit" in dcm else None)
        self.Origin.append([float(pos) for pos in dcm.ImagePositionPatient] if "ImagePositionPatient" in dcm else None)
        self.FilePaths.append(file)

        if "PixelSpacing" in dcm:
            try:
                self.VoxelData.append(
                    VoxelData(x=float(dcm.PixelSpacing[1]), y=float(dcm.PixelSpacing[0]), z=float(dcm.SliceThickness))
                )
            except:
                logger.error("Failed to set VoxelData for dose matrix", exc_info=True)
                self.VoxelData.append(None)

        # Remove pixel data part of dcm to decrease memory used for the object
        if "PixelData" in dcm:
            try:
                del dcm[0x7FE00010]
            except Exception:
                logger.debug(
                    "Failed to remove pixel data from file before appending to CompleteMetadata", exc_info=True
                )
                pass

        self.CompleteMetadata.append(dcm)

    def import_image_volume(self) -> None:
        """Import the file/-s in the Dose matrix volume and insert the data into the ImageVolume property

        Returns:

        """
        for ind, fp in enumerate(self.FilePaths):
            dcm = pydicom.dcmread(fp)

            self.ImageVolume.append(get_pixel_array(dcm=dcm))

    def import_dose_matrix(self) -> None:
        """Import the file/-s in the Dose matrix volume and insert the data into the ImageVolume property

        See :func:`~dicom_image_tools.dicom_handlers.dose_matrix.DoseMatrix.import_image_volume`
        """
        self.import_image_volume()



import logging
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import pydicom
from pydicom import FileDataset

from ..helpers.check_path_is_valid import check_path_is_valid_path
from ..helpers.pixel_data import get_pixel_array
from ..helpers.voxel_data import VoxelData
from .dicom_series import DicomSeries

logger = logging.getLogger(__name__)


class ProjectionSeries(DicomSeries):
    """A class to manage 2D DICOM images, e.g., from conventional X-rays, mammography, panoramic, intraoral etc.

    This class only handles one image per series, as is usually the case for this kind of images.

    Args:
        file: Path object for the file that is to be imported
        dcm: A pydicom FileDataset object containing the file

    Attributes:
        kV: Tube voltage used in the image acquisition in kV
        mA: Tube current used in the image acquisition in mA
        Modality: The modality of the image, e.g., IO, PX, MG, DX, CR, etc.
        Manufacturer: The name of the manufacturer as specified in the DICOM file
        ManufacturersModelName: The model name specified by the manufacturer as given in the DICOM file

    """

    def __init__(self, file: Path, dcm: Optional[FileDataset] = None):
        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if "SeriesInstanceUID" not in dcm:
            raise ValueError("The DICOM file does not contain a series instance UID")

        super().__init__(series_instance_uid=dcm.SeriesInstanceUID)

        self.Modality = dcm.Modality
        self.Manufacturer: Optional[str] = None
        if "Manufacturer" in dcm:
            self.Manufacturer = dcm.Manufacturer

        self.ManufacturersModelName: Optional[str] = None
        if "ManufacturerModelName" in dcm:
            self.ManufacturersModelName = dcm.ManufacturerModelName
        if self.Modality == "IO" and "DetectorManufacturerModelName" in dcm:
            self.ManufacturersModelName = dcm.DetectorManufacturerModelName

        self.kV: Optional[List[Optional[float]]] = []
        self.mA: Optional[List[Optional[float]]] = []
        self.ms: Optional[List[Optional[float]]] = []
        self.ImageVolume: Optional[List[np.ndarray]] = []

        self.add_file(file=file, dcm=dcm)

    def add_file(self, file: Union[Path, str], dcm: Optional[FileDataset] = None):
        """Add a file to the objects list of files

        First performs a check that the file path is of a path object and that it has the same series instance UID as
        the class object

        Args:
            file: Path to where the file to be added is stored on disc
            dcm: The DICOM-file imported to a FileDataset object

        Raises:
            InvalidDicomError: If the given file is not a valid DICOM file
            ValueError: If the file does not have the same study instance UID as the StudyInstanceUID of the object

        """
        file = check_path_is_valid_path(path_to_check=file)

        super().add_file(file=file, dcm=dcm)

        if len(self.CompleteMetadata) == len(self.FilePaths):
            # Skip out because the file has already been added
            return

        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if "PixelSpacing" in dcm:
            self.VoxelData.append(VoxelData(x=float(dcm.PixelSpacing[1]), y=float(dcm.PixelSpacing[0]), z=None))
        else:
            # Assume pixel size is set in Detector Element Spacing tag (0018, 7022)
            self.VoxelData.append(
                VoxelData(x=float(dcm.DetectorElementSpacing[1]), y=float(dcm.DetectorElementSpacing[0]), z=None)
            )

        self.kV.append(self._get_tag_value_as_float_or_none("KVP", ds=dcm))

        if "XRayTubeCurrent" in dcm:
            self.mA.append(float(dcm.XRayTubeCurrent))
        elif "XRayTubeCurrentInmA" in dcm:
            self.mA.append(float(dcm.XRayTubeCurrentInmA))
        elif "XRayTubeCurrentInuA" in dcm:
            self.mA.append(float(dcm.XRayTubeCurrentInuA) / 1000)
        else:
            self.mA.append(None)

        self.ms.append(self._get_tag_value_as_float_or_none("ExposureTime", ds=dcm))

        # Remove pixel data part of dcm to decrease memory used for the object
        if "PixelData" in dcm:
            try:
                del dcm[0x7FE00010]
            except Exception:
                logger.debug(
                    "Failed to remove pixel data from file before appending to CompleteMetadata", exc_info=True
                )
                pass

        self.CompleteMetadata.append(dcm)

    @staticmethod
    def _get_tag_value_as_float_or_none(tag_name: str, ds: FileDataset) -> Optional[float]:
        tag_value = ds.get(tag_name)

        if tag_value is None:
            return None

        return float(tag_value)

    def import_image(self) -> None:
        """Import the pixel data into the ImageVolume property

        Returns:

        """
        self.ImageVolume = []
        for fp in self.FilePaths:
            dcm = pydicom.dcmread(str(fp.absolute()))
            self.ImageVolume.append(get_pixel_array(dcm=dcm))

        if self.PixelIntensityNormalized:
            self.normalize_pixel_intensity_relationship()

    def sort_images_on_acquisition_time(self) -> None:
        """Reorder the images in the series based on the acquisition time

        Returns:

        """
        file_order = [
            ind for ind in list(np.argsort(np.array([self._get_acquisition_time(ds) for ds in self.CompleteMetadata])))
        ]

        self.kV = self._reorder_list_by_index_order_list(order_list=file_order, list_to_order=self.kV)
        self.mA = self._reorder_list_by_index_order_list(order_list=file_order, list_to_order=self.mA)
        self.ms = self._reorder_list_by_index_order_list(order_list=file_order, list_to_order=self.ms)
        self.CompleteMetadata = self._reorder_list_by_index_order_list(
            order_list=file_order, list_to_order=self.CompleteMetadata
        )
        self.FilePaths = self._reorder_list_by_index_order_list(order_list=file_order, list_to_order=self.FilePaths)
        self.ImageVolume = self._reorder_list_by_index_order_list(order_list=file_order, list_to_order=self.ImageVolume)

    @staticmethod
    def _get_acquisition_time(dataset: FileDataset) -> Optional[datetime]:
        acquisition_date = dataset.AcquisitionDate
        acquisition_time = dataset.AcquisitionTime

        acquisition_datetime = f"{acquisition_date} {acquisition_time}"
        datetime_format = "%Y%m%d %H%M%S"

        if "." in acquisition_time:
            datetime_format += ".%f"

        return datetime.strptime(acquisition_datetime, datetime_format)

    @staticmethod
    def _reorder_list_by_index_order_list(order_list: List[int], list_to_order: List[Any]):
        if len(list_to_order) == 0:
            return list_to_order
        return [list_to_order[ind] for ind in order_list]



from dataclasses import dataclass
from typing import Optional


@dataclass
class PatientMassCenter:
    x: float
    y: float
    z: Optional[float] = None


@dataclass
class PatientGeometricalOffset:
    x: float
    y: float



from dataclasses import dataclass
from typing import Optional


@dataclass
class VoxelData:
    """A class for managing voxel/pixel data for DICOM images

    Attributes:
        x: The voxel/pixel x-dimension in mm
        y: The voxel/pixel y-dimension in mm
        z: The voxel/pixel z-dimension in mm
    """

    x: float
    y: float
    z: Optional[float] = None



import logging

import numpy as np
from pydicom import Dataset

log = logging.getLogger(__name__)


def get_pixel_array(dcm: Dataset) -> np.ndarray:
    """Take a DICOM dataset, extract the pixels, rescale the values according to metadata. Return the extracted image
    as a numpy ndarray of int16 values

    Args:
        dcm: The DICOM dataset from which the image should be extracted

    Returns:
        Extracted image as a numpy ndarray of int16 numbers

    """
    px = dcm.pixel_array.astype(np.float)

    if "Modality" in dcm and dcm.Modality.casefold() == "RTDOSE".casefold():
        return rescale_dose_matrix_pixel_array(pixel_array=px, dcm=dcm)

    if "RescaleSlope" in dcm:
        log.debug("Rescaling slope of pixel array")
        px *= np.float(dcm.RescaleSlope)

    if "RescaleIntercept" in dcm:
        log.debug("Rescaling intercept of pixel array")
        px += np.float(dcm.RescaleIntercept)

    return px


def rescale_dose_matrix_pixel_array(pixel_array: np.ndarray, dcm: Dataset) -> np.ndarray:
    """ Rescaled the dose matrix by the DoseGridScaling tag value if it is present in the dataset

    Args:
        pixel_array: The dose matrix as a numpy.ndarray
        dcm: The dataset containing information on the dose matrix

    Returns:
        The rescaled dose matrix as a numpy.ndarray of int16 numbers

    """
    if "DoseGridScaling" in dcm:
        scaling_factor = dcm.DoseGridScaling
        log.debug(f"Rescaling dose matrix by DoseGridScaling of {scaling_factor}")
        pixel_array *= np.int16(scaling_factor)

    return pixel_array



from typing import Dict

from pydicom import FileDataset

from ..constants.SopClassUids import (
    RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS,
    SECONDARY_CAPTURE_SOP_CLASS_UIDS,
)


class DoseReport:
    def __init__(self):
        self.Rdsr: Dict[str, FileDataset] = {}
        self.SecondaryCapture: Dict[str, FileDataset] = {}

    def add_file(self, dataset: FileDataset):
        if not isinstance(dataset, FileDataset):
            raise TypeError("The dataset is not a FileDataset")

        if (
            dataset.SOPClassUID not in RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS
            and dataset.SOPClassUID not in SECONDARY_CAPTURE_SOP_CLASS_UIDS
        ):
            raise ValueError("The supplied FileDataset is neither an RDSR nor a Secondary Capture")

        if dataset.SOPClassUID in RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS:
            self.Rdsr[
                (
                    f"{dataset.ContentDate if 'ContentDate' in dataset else ''}_{dataset.SOPInstanceUID}"
                    f"{dataset.ContentTime if 'ContentTime' in dataset else ''}_{dataset.SOPInstanceUID}"
                )
            ] = dataset

            return

        if dataset.SOPClassUID in SECONDARY_CAPTURE_SOP_CLASS_UIDS:
            self.SecondaryCapture[
                (
                    f"{dataset.ContentDate if 'ContentDate' in dataset else ''}_{dataset.SOPInstanceUID}"
                    f"{dataset.ContentTime if 'ContentTime' in dataset else ''}_{dataset.SOPInstanceUID}"
                )
            ] = dataset



from pathlib import Path
from typing import Optional, Union


def check_path_is_valid_path(path_to_check: Union[str, Path]) -> Optional[Path]:
    """Checks a path given as a string or Path object to see if it is a valid and existing path

    Args:
        path_to_check: Path to validate

    Returns:
        The filepath as a Path-instance if valid file else None

    Raises:
        TypeError: If filepath is neither a string nor a Path
    """
    if isinstance(path_to_check, str):
        path_to_check = Path(path_to_check)

    if not isinstance(path_to_check, Path) or not path_to_check.exists():
        raise TypeError("Invalid path")

    return path_to_check



from pathlib import Path

from setuptools import find_packages, setup

README = (Path(__file__).parent / "README.md").read_text()

setup(
    name="dicom_image_tools",
    version="21.10.3",
    description="Python package for managing DICOM images from different modalities",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/BwKodex/dicomimagetools",
    author="Josef Lundman",
    author_email="josef@lundman.eu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=["pydicom>=2.1.1", "numpy>=1.20.0", "scikit-image>=0.17.2", "scipy>=1.5.4", "SimpleITK>=2.1.1"],
    zip_safe=False,
)



import json
import logging
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from skimage.transform import hough_line, radon

from ..dicom_handlers.ct import CtSeries
from ..helpers.voxel_data import VoxelData

log = logging.getLogger(__name__)


@dataclass
class EquivalentDiameterData:
    Area_px: float
    EAD_px: float
    EAD_cm: float
    EquivalentAreaCircumference_cm: float
    MeanHU: float
    MedianHU: float
    LAT_cm: float
    AP_cm: float
    WED_px: float
    WED_cm: float
    EED_px: float
    EED_cm: float

    def to_dict(self):
        return {
            "Area_px": self.Area_px,
            "EAD_px": self.EAD_px,
            "EAD_cm": self.EAD_cm,
            "EquivalentAreaCircumference_cm": self.EquivalentAreaCircumference_cm,
            "MeanHU": self.MeanHU,
            "MedianHU": self.MedianHU,
            "LAT_cm": self.LAT_cm,
            "AP_cm": self.AP_cm,
            "WED_px": self.WED_px,
            "WED_cm": self.WED_cm,
            "EED_px": self.EED_px,
            "EED_cm": self.EED_cm,
        }

    def to_json(self):
        return json.dumps(self.to_dict())


@dataclass
class SinogramData:
    MaxPixels: float
    MinPixels: float
    MaxCm: float
    MinCm: float
    MaxAngle: float
    MinAngle: float
    EquivalentDiameter: float

    def to_dict(self):
        return {
            "MaxPixels": self.MaxPixels,
            "MinPixels": self.MinPixels,
            "MaxCm": self.MaxCm,
            "MinCm": self.MinCm,
            "MaxAngle": self.MaxAngle,
            "MinAngle": self.MinAngle,
        }

    def to_json(self):
        return json.dumps(self.to_dict())


def _calculate_slice_area_equivalent_diameter(
    image: np.ndarray, mask: np.ndarray, voxel_data: VoxelData
) -> EquivalentDiameterData:
    """Calculate the area equivalent diameter from a 2D image based on AAPM report 204 and 220. Return an
    EquivalentDiameterData instance.

    Args:
        image: A 2D numpy.ndarray containing the patient image
        mask: A 2D binary numpy.ndarray containing the patient mask/ROI
        voxel_data: Pixel size for the mask

    Returns:
        The patient area in pixels, equivalent area diameter (EAD), water equivalent diameter (WED), and elliptical
        equivalent diameter (EED) in pixels and cm, the equivalent area circumference in cm, mean and median
        CT-number/HU, and the lateral and anterio-posterior dimension in cm.

    """
    if not isinstance(image, np.ndarray):
        raise TypeError("image must be a numpy ndarray")

    if not isinstance(mask, np.ndarray):
        raise TypeError("mask must be a numpy ndarray")

    if not isinstance(voxel_data, VoxelData):
        raise TypeError("voxel_data must be a VoxelData instance")

    im = image.copy()  # Prevent overwriting image

    mask_area_pixels = float(np.sum(mask))

    mask_equivalent_diameter_pixels = np.multiply(2.0, np.sqrt(np.divide(mask_area_pixels, np.pi)))
    mask_equivalent_diameter_cm = np.multiply(mask_equivalent_diameter_pixels, np.divide(voxel_data.x, 10.0))
    mask_equivalent_circumference_pixels = np.multiply(2.0, np.sqrt(np.multiply(mask_area_pixels, np.pi)))
    mask_equivalent_circumference_cm = np.multiply(mask_equivalent_circumference_pixels, np.divide(voxel_data.x, 10.0))

    mean_hu = float(np.mean(im[mask]))
    median_hu = float(np.median(im[mask]))

    # Aw = 1/1000 * AvgCT*Aroi + Aroi  from AAPM Report 220 §2.1
    # Aw = sum(CT(x,y))/1000 * Apixel + sum(Apixel)  from AAPM Report 220 §2.1
    # -> = sum(CT(x,y))/1000 * Apixel + Npixel*Apixel
    # -> = (sum(CT(x,y))/1000 + Npixel) * Apixel   # Apixel = area of pixel = pixelspacing[0]*pixelspacing[1]
    aw_pixels = np.add(np.multiply(np.divide(mean_hu, 1000.0), mask_area_pixels), mask_area_pixels)
    wed_pixels = np.multiply(2.0, np.sqrt(np.divide(aw_pixels, np.pi)))
    wed_cm = np.multiply(wed_pixels, np.divide(voxel_data.x, 10.0))

    # Calculate AP and LAT dimensions
    lat_pixels = np.count_nonzero(np.sum(mask, axis=0))
    lat_cm = np.multiply(lat_pixels, np.divide(voxel_data.x, 10.0))
    ap_pixels = np.count_nonzero(np.sum(mask, axis=1))
    ap_cm = np.multiply(ap_pixels, np.divide(voxel_data.y, 10.0))

    # Elliptical efficient diameter (EED) as suggested in AAPM report 204
    # Calculated from the AP and LAT dimensions
    eed_pixels = np.sqrt(np.multiply(ap_pixels, lat_pixels))
    eed_cm = np.sqrt(np.multiply(ap_cm, lat_cm))

    return EquivalentDiameterData(
        Area_px=mask_area_pixels,
        EAD_px=mask_equivalent_diameter_pixels,
        EAD_cm=mask_equivalent_diameter_cm,
        EquivalentAreaCircumference_cm=mask_equivalent_circumference_cm,
        MeanHU=mean_hu,
        MedianHU=median_hu,
        LAT_cm=lat_cm,
        AP_cm=ap_cm,
        WED_px=wed_pixels,
        WED_cm=wed_cm,
        EED_px=eed_pixels,
        EED_cm=eed_cm,
    )


def calculate_max_min_lat_ap_hough(mask: np.ndarray, voxel_data: VoxelData):
    """Calculate the maximum and minimum distance through the patient/mask, and the angles at which they occur, using
    the Hough transform.

    Args:
        mask: A 2D binary numpy.ndarray containing the patient mask/ROI
        voxel_data: Pixel size for the mask

    Returns:
        The maximum and minimum distance through the patient/mask in pixels and cm, and the angles of the maximum and
        minimum distance, respectively.
    """
    if not isinstance(mask, np.ndarray) or (len(mask.shape) > 2 and mask.shape[2] > 1):
        raise TypeError("mask must be a 2D numpy ndarray")

    if not isinstance(voxel_data, VoxelData):
        raise TypeError("voxel_data must be a VoxelData object instance")

    h, angle, f = hough_line(mask, np.linspace(0, np.pi - np.divide(np.pi, 180), 180))

    # Calculate number of non-zero elements for each Hough transform line
    non_zero = [np.count_nonzero(h[:, ind]) for ind in range(0, 180)]

    # Get the maximum size of the patient from the sinogram
    sinogram_max_px = max(non_zero)
    sinogram_max_cm = np.multiply(sinogram_max_px, np.divide(voxel_data.x, 10.0))  # Assumes max patient size = LAT

    # Get the angle of the sinogram_max_px
    max_indices = np.where(np.array(non_zero) == sinogram_max_px)
    min_diffs = [
        min(angle[max_indices]),
        min([np.absolute(ang - np.divide(np.pi, 2.0)) for ang in angle[max_indices]]),
        min([np.absolute(ang - np.pi) for ang in angle[max_indices]]),
    ]
    tmp_min_ind = min_diffs.index(min(min_diffs))
    if tmp_min_ind == 0:
        sinogram_max_degree = [ang for ang in angle[max_indices] if ang == min_diffs[tmp_min_ind]][0]
    elif tmp_min_ind == 1:
        sinogram_max_degree = [
            ang for ang in angle[max_indices] if np.absolute(ang - np.divide(np.pi, 2.0)) == min_diffs[tmp_min_ind]
        ][0]
    else:
        sinogram_max_degree = [ang for ang in angle[max_indices] if np.absolute(ang - np.pi) == min_diffs[tmp_min_ind]][
            0
        ]

    sinogram_max_degree = np.multiply(sinogram_max_degree, np.divide(180.0, np.pi))

    # Get the minimum size of the patient from the sinogram
    sinogram_min_px = min(non_zero)
    sinogram_min_cm = np.multiply(sinogram_min_px, np.divide(voxel_data.y, 10.0))  # Assumes min patient size = AP

    min_indices = np.where(np.array(non_zero) == sinogram_min_px)
    min_diffs = [
        min(angle[min_indices]),
        min([np.absolute(ang - np.divide(np.pi, 2.0)) for ang in angle[min_indices]]),
        min([np.absolute(ang - np.pi) for ang in angle[min_indices]]),
    ]
    tmp_min_ind = min_diffs.index(min(min_diffs))
    if tmp_min_ind == 0:
        sinogram_min_degree = [ang for ang in angle[min_indices] if ang == min_diffs[tmp_min_ind]][0]
    elif tmp_min_ind == 1:
        sinogram_min_degree = [
            ang for ang in angle[min_indices] if np.absolute(ang - np.divide(np.pi, 2.0)) == min_diffs[tmp_min_ind]
        ][0]
    else:
        sinogram_min_degree = [ang for ang in angle[min_indices] if np.absolute(ang - np.pi) == min_diffs[tmp_min_ind]][
            0
        ]
    sinogram_min_degree = np.multiply(sinogram_min_degree, np.divide(180.0, np.pi))

    return SinogramData(
        MaxPixels=sinogram_max_px,
        MaxCm=sinogram_max_cm,
        MaxAngle=sinogram_max_degree,
        MinPixels=sinogram_min_px,
        MinCm=sinogram_min_cm,
        MinAngle=sinogram_min_degree,
        EquivalentDiameter=np.sqrt(np.multiply(sinogram_max_cm, sinogram_min_cm)),
    )


def calculate_max_min_lat_ap_radon(mask: np.ndarray, voxel_data: VoxelData) -> SinogramData:
    """Calculate the maximum and minimum distance through the patient/mask, and the angles at which they occur, using
    the radon transform.

    Args:
        mask: A 2D binary numpy.ndarray containing the patient mask/ROI
        voxel_data: Pixel size for the mask

    Returns:
        The maximum and minimum distance through the patient/mask in pixels and cm, and the angles of the maximum and
        minimum distance, respectively.
    """
    if not isinstance(mask, np.ndarray) or (len(mask.shape) > 2 and mask.shape[2] > 1):
        raise TypeError("mask must be a 2D numpy ndarray")

    if not isinstance(voxel_data, VoxelData):
        raise TypeError("voxel_data must be a VoxelData object instance")

    radon_mask = radon(mask, theta=np.arange(180), circle=True, preserve_range=True)

    non_zero = [np.count_nonzero(radon_mask[:, ind]) for ind in range(180)]

    sinogram_max_px = max(non_zero)
    sinogram_max_cm = np.multiply(sinogram_max_px, np.divide(voxel_data.x, 10.0))  # Assumes max patient size = LAT

    # Get the angle of the sinogram_max_px
    max_indices = list(np.where(np.array(non_zero) == sinogram_max_px)[0])
    min_diffs = [
        min(max_indices),
        min([np.absolute(ang - 90.0) for ang in max_indices]),
        min([np.absolute(ang - 180.0) for ang in max_indices]),
    ]
    tmp_min_ind = min_diffs.index(min(min_diffs))
    if tmp_min_ind == 0:
        sinogram_max_degree = float([ang for ang in max_indices if ang == min_diffs[tmp_min_ind]][0])
    elif tmp_min_ind == 1:
        sinogram_max_degree = float(
            [ang for ang in max_indices if np.absolute(ang - 90.0) == min_diffs[tmp_min_ind]][0]
        )
    else:
        sinogram_max_degree = float(
            [ang for ang in max_indices if np.absolute(ang - 180.0) == min_diffs[tmp_min_ind]][0]
        )

    sinogram_min_px = min(non_zero)
    sinogram_min_cm = np.multiply(sinogram_min_px, np.divide(voxel_data.y, 10.0))  # Assumes min patient size = AP

    min_indices = list(np.where(np.array(non_zero) == sinogram_min_px)[0])
    min_diffs = [
        min(min_indices),
        min([np.absolute(ang - 90.0) for ang in min_indices]),
        min([np.absolute(ang - 180.0) for ang in min_indices]),
    ]
    tmp_min_ind = min_diffs.index(min(min_diffs))
    if tmp_min_ind == 0:
        sinogram_min_degree = float([ang for ang in min_indices if ang == min_diffs[tmp_min_ind]][0])
    elif tmp_min_ind == 1:
        sinogram_min_degree = float(
            [ang for ang in min_indices if np.absolute(ang - 90.0) == min_diffs[tmp_min_ind]][0]
        )
    else:
        sinogram_min_degree = float(
            [ang for ang in min_indices if np.absolute(ang - 180.0) == min_diffs[tmp_min_ind]][0]
        )

    return SinogramData(
        MaxPixels=sinogram_max_px,
        MaxCm=sinogram_max_cm,
        MaxAngle=sinogram_max_degree,
        MinPixels=sinogram_min_px,
        MinCm=sinogram_min_cm,
        MinAngle=sinogram_min_degree,
        EquivalentDiameter=np.sqrt(np.multiply(sinogram_max_cm, sinogram_min_cm)),
    )


def calculate_area_equivalent_diameter(
    ct: CtSeries, use_radon: Optional[bool] = False, use_hough: Optional[bool] = False
) -> Tuple[List[EquivalentDiameterData], Optional[List[SinogramData]], Optional[List[SinogramData]]]:
    """Calculate the area equivalent diameter for all slices in a CT series. The calculations are based on the
    specifications in AAPM report 204 and 220.

    Args:
        ct: A CtSeries object containing the series for which the area equivalent diameter should be determined
        use_radon: Use radon transform to determine maximum and minimum distance through patient
        use_hough: Use Hough transform to determine maximum and minimum distance through patient

    Returns:
        Tuple in the form (EAD_data, radon_data, hough_data) where the radon_data and hough_data only have values if
        use_radon and/or use_hough was set to true, respectively.
    """
    if not isinstance(ct, CtSeries):
        raise TypeError("ct must be an instance of the CtSeries class")

    if not isinstance(use_radon, bool):
        raise TypeError("use_radon must be a boolean")

    if not isinstance(use_hough, bool):
        raise TypeError("use_hough must be a boolean")

    log.debug("Calculating area equivalent diameter")
    if ct.Mask is None:
        ct.get_patient_mask(remove_table=True)

    ead_data = [
        _calculate_slice_area_equivalent_diameter(ct.ImageVolume[:, :, ind], ct.Mask[:, :, ind], ct.VoxelData[ind])
        for ind in range(ct.ImageVolume.shape[2])
    ]

    radon_data = None
    if use_radon:
        radon_data = [
            calculate_max_min_lat_ap_radon(mask=ct.Mask[:, :, ind], voxel_data=ct.VoxelData[ind])
            for ind in range(ct.ImageVolume.shape[2])
        ]

    hough_data = None
    if use_hough:
        hough_data = [
            calculate_max_min_lat_ap_radon(mask=ct.Mask[:, :, ind], voxel_data=ct.VoxelData[ind])
            for ind in range(ct.ImageVolume.shape[2])
        ]

    return ead_data, radon_data, hough_data
